#include <propeller2.h>
#include "common.h"
#include "hal.h"
#include "config.h"

struct exmem_mailbox {
    uint32_t cmd;
    void *hubPtr;
    uint32_t length;
};

static volatile struct exmem_mailbox exmem_mailbox[8];
static uint32_t exmem_size;
static uint32_t exmem_cfg[8],exmem_cogs[8],exmem_banks[32];

static struct __using(
#if P2_EXMEM_TYPE == EXMEM_PSRAM16
    "psram16drv.spin2"
#define EXMEM_IS_PSRAM
#elif P2_EXMEM_TYPE == EXMEM_PSRAM8
    "psram8drv.spin2"
#define EXMEM_IS_PSRAM
#elif P2_EXMEM_TYPE == EXMEM_PSRAM4
    "psram4drv-dualCE.spin2"
#define EXMEM_IS_PSRAM
#elif P2_EXMEM_TYPE == EXMEM_HYPERRAM
    "hyperdrv.spin2"
#define EXMEM_IS_HYPER
#endif
) exmem_driver;

static struct __using("ExVGA.spin2") VGA;

static volatile uint32_t lineCtr, frameCtr, swapRequest;
static uint16_t lineBuffers[320*2];

static uint8_t exmem_cog,video_cog;
static uint8_t buffer_flag;

static struct __using("usbnew.spin2", USB_BASE_PIN = P2_USB_BASEPIN, USB_ENABLE_OFFSET = P2_USB_ENABLE_OFFSET, USB_DMINUS_OFFSET = P2_USB_DM_OFFSET) USB;

void HAL_startup() {


    // Video startup (sets clock!)
    {
        // WARNING: We assume video cog ends up as cog 1
        video_cog = 1;
        VGA.start(&lineCtr,&exmem_mailbox[video_cog],lineBuffers,VGA.P2_VIDEO_MODE,P2_VIDEO_BASE,P2_VIDEO_VSYNC,P2_VIDEO_SUBMODE);
    }

    // External memory startup
    {
        int sizelen;
        #ifdef EXMEM_IS_HYPER
            exmem_cfg[1] = (P2_EXMEM_SYNC_CLOCK?0u:1u)<<exmem_driver.UNREGCLK_BIT;
            exmem_cfg[2] = P2_EXMEM_RESET < 32 ? 1u<<P2_EXMEM_RESET : 0;
            exmem_cfg[3] = P2_EXMEM_RESET >= 32 ? 1u<<P2_EXMEM_RESET : 0;
            exmem_cfg[4] = P2_EXMEM_BASE;
        #endif
        #ifdef EXMEM_IS_PSRAM
            exmem_cfg[1] = (P2_EXMEM_SYNC_CLOCK?0u:1u)<<exmem_driver.UNREGCLK_BIT;
            #if P2_EXMEM_SLOW
                exmem_cfg[1] |= 1<<exmem_driver.SLOWCLK_BIT | 1<<exmem_driver.CLKSEL_BIT;
            #endif
            exmem_cfg[2] = 0;
            exmem_cfg[3] = 0;
            exmem_cfg[4] = P2_EXMEM_BASE;
        #endif
        exmem_cfg[5] = (uint32_t)exmem_banks;
        exmem_cfg[6] = (uint32_t)exmem_cogs;
        exmem_cfg[7] = (uint32_t)exmem_mailbox;
        
        longset(exmem_cogs,-1<<16,8);
        longset(exmem_banks,INT32_MIN,32);

        // Set QoS for video
        exmem_cogs[video_cog] = (-1<<16)+(15<<12)+(0<<10)+(0<<11);

        #if P2_EXMEM_TYPE == EXMEM_PSRAM16
            exmem_size = 32*1024*1024 * P2_EXMEM_BANKS;
            sizelen = 33-__builtin_clz(exmem_size-1);
            for(int i=0;i<P2_EXMEM_BANKS;i++) {
                exmem_banks[0+i*2] = (2048<<16) + (P2_EXMEM_DELAY<<12) + sizelen;
                exmem_banks[1+i*2] = (2048<<16) + (P2_EXMEM_DELAY<<12) + sizelen;
                exmem_banks[16+i*2] = (P2_EXMEM_SELECT+i) + (P2_EXMEM_CLK<<8);
                exmem_banks[17+i*2] = (P2_EXMEM_SELECT+i) + (P2_EXMEM_CLK<<8);
            }
        #elif P2_EXMEM_TYPE == EXMEM_PSRAM8
            exmem_size = 16*1024*1024 * P2_EXMEM_BANKS;
            sizelen = 33-__builtin_clz(exmem_size-1);
            for(int i=0;i<P2_EXMEM_BANKS;i++) {
                exmem_banks[0+i] = (1024<<16) + (P2_EXMEM_DELAY<<12) + sizelen;
                exmem_banks[16+i] = (P2_EXMEM_SELECT+i) + (P2_EXMEM_CLK<<8);
            }
        #elif P2_EXMEM_TYPE == EXMEM_PSRAM4
            exmem_size = 8*1024*1024 * P2_EXMEM_BANKS;
            sizelen = 33-__builtin_clz(exmem_size-1);
            for(int i=0;i<(P2_EXMEM_BANKS+1)/2;i++) {
                exmem_banks[0+i] = (512<<16) + (P2_EXMEM_DELAY<<12) + sizelen;
                exmem_banks[16+i] = (P2_EXMEM_SELECT+i*2) + (P2_EXMEM_CLK<<8) + ((P2_EXMEM_SELECT+i*2+(((P2_EXMEM_BANKS&1)&&(i>=P2_EXMEM_BANKS/2)) ? 0 : 1))<<16);
            }
        #elif P2_EXMEM_TYPE == EXMEM_HYPERRAM
            exmem_size = 16*1024*1024 * P2_EXMEM_BANKS;
            sizelen = 33-__builtin_clz(exmem_size-1);
            for(int i=0;i<P2_EXMEM_BANKS;i++) {
                exmem_banks[0+i] := (1024<<16) + (P2_EXMEM_DELAY<<12) + sizelen;
                exmem_banks[16+i] := (P2_EXMEM_SELECT+i) + (P2_EXMEM_CLK<<8) + (P2_EXMEM_RWDS<<16) + (6<<25); // Latency????
            }
        #endif

        exmem_mailbox[0].cmd = UINT32_MAX;
        exmem_cog = _coginit(COGEXEC_NEW,exmem_driver.getDriverAddr(),exmem_cfg);
        while (exmem_mailbox[0].cmd) {} // wait for init
        HAL_ex_fill(0,0,exmem_size,true);
    }

    // USB startup
    {
        USB.start();
    }
    
}

bool HAL_keystate(int scancode) {
    return USB.keystate(scancode);
}

void HAL_submit_frame(void *frame) {
    // The +1 -1 thing bypasses an awful flexspin bug
    uint32_t base = exmem_size-(320*240*2*2)+((buffer_flag+1)-1)*(320*240*2);
    while(swapRequest) {}
    HAL_ex_write(base,frame,320*240*2,true);
}

void HAL_flip_buffers() {
    while(swapRequest) {}
    HAL_ex_sync();
    // The +1 -1 thing bypasses an awful flexspin bug
    swapRequest = exmem_driver.R_READBURST+exmem_size-(320*240*2*2)+((buffer_flag+1)-1)*(320*240*2);
    buffer_flag ^= 1;
}

void HAL_poll_inputs() {

}

uint32_t HAL_ex_size() {
    return exmem_size - 320*240*2*2;
}
static inline volatile struct exmem_mailbox *get_mb() {
    volatile struct exmem_mailbox *mb = &exmem_mailbox[_cogid()];
    while (mb->cmd) {};
    return mb;
}
void HAL_ex_write(uint32_t dst,void *src,uint32_t len,bool async) {
    volatile struct exmem_mailbox *mb = get_mb();
    mb->length = len;
    mb->hubPtr = src;
    mb->cmd = exmem_driver.R_WRITEBURST + (dst & 0xfffffff);
    if(!async) while (mb->cmd) {};
}
void HAL_ex_fill(uint32_t dst,int value,uint32_t len,bool async) {
    volatile struct exmem_mailbox *mb = get_mb();
    mb->length = len;
    mb->hubPtr = (void *)value;
    mb->cmd = exmem_driver.R_WRITEBYTE + (dst & 0xfffffff);
    if(!async) while (mb->cmd) {};
}
void HAL_ex_read(void *dst, uint32_t src,uint32_t len,bool async) {
    volatile struct exmem_mailbox *mb = get_mb();
    mb->length = len;
    mb->hubPtr = dst;
    mb->cmd = exmem_driver.R_READBURST + (src & 0xfffffff);
    if(!async) while (mb->cmd) {};
}
void HAL_ex_sync() {
    get_mb();
}
