
#include <SDL2/SDL.h>
#include "common.h"
#include "hal.h"

static SDL_Window *window;
static SDL_Renderer *render;
static SDL_Texture *screenTexture;

static const uint8_t *keystates;
static int max_scancode = 0;

__attribute__((format (printf,1,2)))
static void doError(const char *f,...) {
    char mesg[1024];
    va_list argptr;
    va_start(argptr,f);
    vsnprintf(mesg,1024,f,argptr);
    va_end(argptr);
    SDL_ShowSimpleMessageBox(SDL_MESSAGEBOX_ERROR,"Error",mesg,window);
    exit(-1);
}

void HAL_startup() {
    if (SDL_Init(SDL_INIT_EVERYTHING)<0) doError("Failed to initialize SDL: %s",SDL_GetError());
    
    window = SDL_CreateWindow(NULL
    , SDL_WINDOWPOS_CENTERED,SDL_WINDOWPOS_CENTERED,640,480,SDL_WINDOW_RESIZABLE|SDL_WINDOW_HIDDEN);
    if (!window) doError("Failed to create window: %s",SDL_GetError());
    SDL_SetWindowTitle(window,"P2craft (SDL2 HAL)");

    render = SDL_CreateRenderer(window,-1,0|SDL_RENDERER_PRESENTVSYNC|SDL_RENDERER_ACCELERATED);
    if (!render) {
        doError("Failed to create renderer: %s",SDL_GetError());
    }
    SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY,"best");

    screenTexture = SDL_CreateTexture(render,SDL_PIXELFORMAT_RGB565,SDL_TEXTUREACCESS_STREAMING,320,240);
    if (!screenTexture) doError("Failed to create screenTexture: %s",SDL_GetError());

    SDL_ShowWindow(window);

    keystates = SDL_GetKeyboardState(&max_scancode);
}

void HAL_poll_inputs() {
    SDL_Event ev;
    while (SDL_PollEvent(&ev)) {
        switch(ev.type) {
        case SDL_QUIT: {
            exit(0);
        } break;
        }
    }
}

bool HAL_keystate(int scancode) {
    if (scancode > max_scancode) return false;
    return keystates[scancode];
}

void HAL_submit_frame(const void *frame) {
    SDL_UpdateTexture(screenTexture,NULL,frame,320*2);
}

void HAL_flip_buffers() {
    int out_w,out_h;
    SDL_GetRendererOutputSize(render,&out_w,&out_h);
    SDL_SetRenderDrawColor(render,0,0,0,255);
    SDL_RenderClear(render);
    float pixel_aspect = 1.0;
    float wratio = ((float)out_w/pixel_aspect)/320;
    float hratio = (float)out_h/240;
    float lbratio = fminf(wratio,hratio);
    SDL_Rect screenDest = {.x=(int)(out_w/2 - lbratio*pixel_aspect*160),.y=(int)(out_h/2 - lbratio*120),.w=(int)(lbratio*pixel_aspect*320),.h=(int)(lbratio*240)};
    SDL_Rect src = {.x=0,.y=0,.w=320,.h=240};
    SDL_RenderCopy(render,screenTexture,&src,&screenDest);
    SDL_RenderPresent(render);
}


#define VIRTUAL_EXMEM_SIZE (32*1024*1024)

static uint8_t external_memory[VIRTUAL_EXMEM_SIZE];

uint32_t HAL_ex_size() {
    return VIRTUAL_EXMEM_SIZE;
}
void HAL_ex_write(uint32_t dst,void *src,uint32_t len,bool async) {
    memcpy(external_memory+dst,src,len);
}
void HAL_ex_fill(uint32_t dst,int value,uint32_t len,bool async) {
    memset(external_memory+dst,value,len);
}
void HAL_ex_read(void *dst, uint32_t src,uint32_t len,bool async) {
    memcpy(dst,external_memory+src,len);
}
void HAL_ex_sync() {
    // No-op
}

